from titus_isolate import log
from titus_isolate.allocate.cpu_allocator import CpuAllocator
from titus_isolate.model.processor.cpu import Cpu


class NoopCpuAllocator(CpuAllocator):
    def __init__(self, free_thread_provider=""):
        pass

    def assign_threads(self, cpu: Cpu, workload_id: str, workloads: dict) -> Cpu:
        log.info("Ignoring attempt to assign threads to workload: '{}'".format(workload_id))

    def free_threads(self, cpu: Cpu, workload_id: str, workloads: dict) -> Cpu:
        log.info("Ignoring attempt to free threads for workload: '{}'".format(workload_id))

    def rebalance(self, cpu: Cpu, workloads: dict) -> Cpu:
        return cpu

    def set_registry(self, registry):
        pass

    def report_metrics(self, tags):
        pass



from abc import abstractmethod


class CpuUsageProvider:

    @abstractmethod
    def get_cpu_usage(self, seconds: int, agg_granularity_secs : int) -> dict:
        """
        Returns CPU usage per workload over the last `seconds` seconds, aggregated at `agg_granularity_secs`.
        """
        pass



from titus_isolate.event.constants import BURST, STATIC
from titus_isolate.model.processor.core import Core
from titus_isolate.model.processor.cpu import Cpu
from titus_isolate.model.processor.thread import Thread
from titus_isolate.model.utils import get_burst_workloads


def get_allocated_size(cpu: Cpu) -> int:
    return len([t for t in cpu.get_threads() if t.is_claimed()])


def get_unallocated_size(cpu: Cpu) -> int:
    return len([t for t in cpu.get_threads() if not t.is_claimed()])


def get_burst_request_size(workloads: list) -> int:
    burst_request_size = 0
    for w in get_burst_workloads(workloads):
        burst_request_size += w.get_thread_count()

    return burst_request_size


def get_burst_allocated_size(cpu: Cpu, workload_map: dict) -> int:
    return _get_allocated_size(cpu, workload_map, BURST)


def get_static_allocated_size(cpu: Cpu, workload_map: dict) -> int:
    return _get_allocated_size(cpu, workload_map, STATIC)


def _get_allocated_size(cpu: Cpu, workload_map: dict, w_type: str) -> int:
    allocation_size = 0
    for t in cpu.get_threads():
        if _is_thread_occupied(t, workload_map, w_type):
            allocation_size += 1

    return allocation_size


def get_oversubscribed_thread_count(cpu: Cpu, workload_map: dict) -> int:
    oversubscribed_thread_count = 0
    for t in cpu.get_threads():
        if _is_thread_occupied(t, workload_map, BURST) and _is_thread_occupied(t, workload_map, STATIC):
            oversubscribed_thread_count += 1

    return oversubscribed_thread_count


def _is_thread_occupied(thread: Thread, workload_map: dict, w_type: str) -> bool:
    for w_id in thread.get_workload_ids():
        if w_id in workload_map:
            if workload_map[w_id].get_type() == w_type:
                return True

    return False



# Metrics
DEFAULT_SAMPLE_FREQUENCY_SEC = 6

# Rebalance
REBALANCE_FREQUENCY_KEY = 'TITUS_ISOLATE_REBALANCE_FREQUENCY'
DEFAULT_REBALANCE_FREQUENCY = 18

# Reconcile
RECONCILE_FREQUENCY_KEY = 'TITUS_ISOLATE_RECONCILE_FREQUENCY'
DEFAULT_RECONCILE_FREQUENCY = 60

# CPU Allocator
CPU_ALLOCATOR = 'TITUS_ISOLATE_ALLOCATOR'
FALLBACK_ALLOCATOR = 'TITUS_ISOLATE_FALLBACK_ALLOCATOR'

# Remote Allocator
REMOTE_ALLOCATOR_URL = 'TITUS_ISOLATE_REMOTE_ALLOCATOR_URL'

AB_TEST = 'AB_TEST'
IP = 'IP'
FORECAST_CPU_IP = 'FORECAST_CPU_IP'
GREEDY = 'GREEDY'
NOOP = 'NOOP'
NOOP_RESET = 'NOOP_RESET'
REMOTE = 'REMOTE'
DEFAULT_ALLOCATOR = IP
DEFAULT_FALLBACK_ALLOCATOR = GREEDY
CPU_ALLOCATORS = [AB_TEST, IP, FORECAST_CPU_IP, GREEDY, NOOP, NOOP_RESET, REMOTE]

CPU_ALLOCATOR_A = 'CPU_ALLOCATOR_A'
CPU_ALLOCATOR_B = 'CPU_ALLOCATOR_B'

# Forecast CPU Allocator
ALPHA_NU = 'TITUS_ISOLATE_ALPHA_NU'
DEFAULT_ALPHA_NU = 1000.0

ALPHA_LLC = 'TITUS_ISOLATE_ALPHA_LLC'
DEFAULT_ALPHA_LLC = 10.0

ALPHA_L12 = 'TITUS_ISOLATE_ALPHA_L12'
DEFAULT_ALPHA_L12 = 250.0

ALPHA_ORDER = 'TITUS_ISOLATE_ALPHA_ORDER'
DEFAULT_ALPHA_ORDER = 1.0

ALPHA_PREV = 'TITUS_ISOLATE_ALPHA_PREV'
DEFAULT_ALPHA_PREV = 10.0

BURST_MULTIPLIER = 'TITUS_ISOLATE_BURST_MULTIPLIER'
DEFAULT_BURST_MULTIPLIER = 0.1

MAX_BURST_POOL_INCREASE_RATIO = 'TITUS_ISOLATE_MAX_BURST_POOL_INCREASE_RATIO'
DEFAULT_MAX_BURST_POOL_INCREASE_RATIO = 3.0

BURST_CORE_COLLOC_USAGE_THRESH = 'TITUS_ISOLATE_BURST_CORE_COLLOC_USAGE_THRESH'
DEFAULT_BURST_CORE_COLLOC_USAGE_THRESH = 0.2

WEIGHT_CPU_USE_BURST = 'TITUS_ISOLATE_WEIGHT_CPU_USE_BURST'
DEFAULT_WEIGHT_CPU_USE_BURST = 1.0

RELATIVE_MIP_GAP_STOP = 'TITUS_RELATIVE_MIP_GAP_STOP'
DEFAULT_RELATIVE_MIP_GAP_STOP = 0.05

MIP_SOLVER = 'TITUS_MIP_SOLVER'
DEFAULT_MIP_SOLVER = 'GLPK_MI'

# Free Thread Provider
FREE_THREAD_PROVIDER = 'FREE_THREAD_PROVIDER'
EMPTY = 'EMPTY'
THRESHOLD = 'THRESHOLD'
DEFAULT_FREE_THREAD_PROVIDER = EMPTY

# Threshold Free Thread Provider
TOTAL_THRESHOLD = 'TOTAL_THRESHOLD'
DEFAULT_TOTAL_THRESHOLD = 0.1

THRESHOLD_TOTAL_DURATION_SEC = 'THRESHOLD_TOTAL_DURATION_SEC'
DEFAULT_THRESHOLD_TOTAL_DURATION_SEC = 600

PER_WORKLOAD_THRESHOLD = 'PER_WORKLOAD_THRESHOLD'
DEFAULT_PER_WORKLOAD_THRESHOLD = 0.05

PER_WORKLOAD_DURATION_SEC = 'PER_WORKLOAD_DURATION_SEC'
DEFAULT_PER_WORKLOAD_DURATION_SEC = DEFAULT_SAMPLE_FREQUENCY_SEC

# cgroup File
WAIT_CGROUP_FILE_KEY = 'TITUS_ISOLATE_WAIT_CGROUP_FILE_SEC'
DEFAULT_WAIT_CGROUP_FILE_SEC = 90

# JSON File
WAIT_JSON_FILE_KEY = 'TITUS_ISOLATE_WAIT_JSON_FILE_SEC'
DEFAULT_WAIT_JSON_FILE_SEC = 10

# Blocking isolation wait
TITUS_ISOLATE_BLOCK_SEC = 'TITUS_ISOLATE_BLOCK_SEC'
DEFAULT_TITUS_ISOLATE_BLOCK_SEC = 10

# NUMA balancing
TITUS_ISOLATE_DYNAMIC_NUMA_BALANCING = 'TITUS_ISOLATE_DYNAMIC_NUMA_BALANCING'
DEFAULT_TITUS_ISOLATE_DYNAMIC_NUMA_BALANCING = True

# Event log
EVENT_LOG_FORMAT_STR = 'EVENT_LOG_FORMAT_STR'

# S3 Buckets
DEV = 'prod'
V1 = 'v1'
LATEST = 'latest'

MODEL_BUCKET_FORMAT_STR = 'TITUS_ISOLATE_MODEL_BUCKET_FORMAT_STR'
MODEL_PREFIX_FORMAT_STR = 'TITUS_ISOLATE_MODEL_PREFIX_FORMAT_STR'

MODEL_BUCKET_PREFIX = 'TITUS_ISOLATE_MODEL_BUCKET_PREFIX'
DEFAULT_MODEL_BUCKET_PREFIX = DEV

MODEL_BUCKET_LEAF = 'TITUS_ISOLATE_MODEL_BUCKET_LEAF'
DEFAULT_MODEL_BUCKET_LEAF = LATEST

# Static environment variables
PROPERTY_URL_ROOT = 'http://localhost:3002/properties'
EC2_INSTANCE_ID = "EC2_INSTANCE_ID"

MAX_SOLVER_RUNTIME = 'TITUS_ISOLATE_MAX_SOLVER_RUNTIME'
DEFAULT_MAX_SOLVER_RUNTIME = 1

RESTART_PROPERTIES = [
    ALPHA_NU,
    ALPHA_LLC,
    ALPHA_L12,
    ALPHA_ORDER,
    ALPHA_PREV,
    BURST_CORE_COLLOC_USAGE_THRESH,
    BURST_MULTIPLIER,
    CPU_ALLOCATOR,
    FALLBACK_ALLOCATOR,
    FREE_THREAD_PROVIDER,
    MAX_BURST_POOL_INCREASE_RATIO,
    MAX_SOLVER_RUNTIME,
    MODEL_BUCKET_FORMAT_STR,
    MODEL_PREFIX_FORMAT_STR,
    PER_WORKLOAD_DURATION_SEC,
    PER_WORKLOAD_THRESHOLD,
    REBALANCE_FREQUENCY_KEY,
    RECONCILE_FREQUENCY_KEY,
    REMOTE_ALLOCATOR_URL,
    THRESHOLD_TOTAL_DURATION_SEC,
    TOTAL_THRESHOLD,
    WEIGHT_CPU_USE_BURST]



class CpuUsageSnapshot:

    def __init__(self, timestamp, rows):
        self.timestamp = timestamp
        self.rows = rows


class CpuUsage:

    def __init__(self, pu_id, user, system):
        self.pu_id = pu_id
        self.user = user
        self.system = system



import copy
from threading import Lock, Thread
import time

from titus_isolate import log
from titus_isolate.allocate.cpu_allocator import CpuAllocator
from titus_isolate.allocate.fall_back_cpu_allocator import FallbackCpuAllocator
from titus_isolate.allocate.noop_allocator import NoopCpuAllocator
from titus_isolate.allocate.noop_reset_allocator import NoopResetCpuAllocator
from titus_isolate.cgroup.cgroup_manager import CgroupManager
from titus_isolate.isolate.detect import get_cross_package_violations, get_shared_core_violations
from titus_isolate.isolate.metrics_utils import get_static_allocated_size, get_burst_allocated_size, \
    get_burst_request_size, get_oversubscribed_thread_count, get_allocated_size, get_unallocated_size
from titus_isolate.isolate.update import get_updates
from titus_isolate.metrics.constants import RUNNING, ADDED_KEY, REMOVED_KEY, SUCCEEDED_KEY, FAILED_KEY, \
    WORKLOAD_COUNT_KEY, ALLOCATOR_CALL_DURATION, PACKAGE_VIOLATIONS_KEY, CORE_VIOLATIONS_KEY, \
    ADDED_TO_FULL_CPU_ERROR_KEY, OVERSUBSCRIBED_THREADS_KEY, \
    STATIC_ALLOCATED_SIZE_KEY, BURST_ALLOCATED_SIZE_KEY, BURST_REQUESTED_SIZE_KEY, ALLOCATED_SIZE_KEY, \
    UNALLOCATED_SIZE_KEY, REBALANCED_KEY
from titus_isolate.metrics.event_log import report_cpu
from titus_isolate.metrics.metrics_reporter import MetricsReporter
from titus_isolate.model.processor.cpu import Cpu
from titus_isolate.model.processor.utils import visualize_cpu_comparison
from titus_isolate.numa.utils import update_numa_balancing


class WorkloadManager(MetricsReporter):

    def __init__(self, cpu: Cpu, cgroup_manager: CgroupManager, cpu_allocator: CpuAllocator):

        self.__reg = None
        self.__lock = Lock()

        self.__cpu_allocator = cpu_allocator

        self.__error_count = 0
        self.__added_count = 0
        self.__removed_count = 0
        self.__rebalanced_count = 0
        self.__added_to_full_cpu_count = 0
        self.__allocator_call_duration_sum_secs = 0

        self.__cpu = cpu
        self.__cgroup_manager = cgroup_manager
        self.__workloads = {}

        log.info("Created workload manager")

    def add_workload(self, workload):
        update_numa_balancing(workload, self.__cpu)

        succeeded = self.__update_workload(self.__add_workload, workload, workload.get_id())
        if succeeded:
            self.__added_count += 1
        else:
            self.__remove_workload(workload.get_id())

    def remove_workload(self, workload_id):
        self.__cgroup_manager.release_cpuset(workload_id)
        self.__update_workload(self.__remove_workload, workload_id, workload_id)
        self.__removed_count += 1

    def rebalance(self):
        with self.__lock:
            log.debug("Rebalancing...")
            new_cpu = self.get_cpu_copy()
            workload_map = self.get_workload_map_copy()
            new_cpu = self.__cpu_allocator.rebalance(new_cpu, workload_map)
            self.__rebalanced_count += 1
            self.__update_state(new_cpu, workload_map)
            log.debug("Rebalanced")

    def __update_workload(self, func, arg, workload_id):
        try:
            with self.__lock:
                log.debug("Acquired lock for func: {} on workload: {}".format(func.__name__, workload_id))
                start_time = time.time()
                func(arg)
                stop_time = time.time()
                self.__allocator_call_duration_sum_secs = stop_time - start_time

            log.debug("Released lock for func: {} on workload: {}".format(func.__name__, workload_id))
            return True
        except:
            self.__error_count += 1
            log.exception("Failed to execute func: {} on workload: {}".format(func.__name__, workload_id))
            return False

    def __add_workload(self, workload):
        log.info("Assigning '{}' thread(s) to workload: '{}'".format(workload.get_thread_count(), workload.get_id()))

        new_cpu = self.get_cpu_copy()
        workload_map = self.get_workload_map_copy()
        workload_map[workload.get_id()] = workload

        new_cpu = self.__cpu_allocator.assign_threads(new_cpu, workload.get_id(), workload_map)
        self.__update_state(new_cpu, workload_map)

    def __remove_workload(self, workload_id):
        log.info("Removing workload: {}".format(workload_id))
        if workload_id not in self.__workloads:
            log.error("Attempted to remove unknown workload: '{}'".format(workload_id))
            return

        new_cpu = self.get_cpu_copy()
        workload_map = self.get_workload_map_copy()
        workload = workload_map[workload_id]

        new_cpu = self.__cpu_allocator.free_threads(new_cpu, workload.get_id(), workload_map)
        workload_map.pop(workload.get_id())
        self.__update_state(new_cpu, workload_map)

    def __update_state(self, new_cpu, new_workloads):
        old_cpu = self.get_cpu_copy()
        updated = self.__apply_cpuset_updates(old_cpu, new_cpu)
        self.__cpu = new_cpu
        self.__workloads = new_workloads

        if updated:
            self.__report_cpu_state(old_cpu, new_cpu)

    def __apply_cpuset_updates(self, old_cpu, new_cpu):
        updates = get_updates(old_cpu, new_cpu)
        for workload_id, thread_ids in updates.items():
            log.info("updating workload: '{}' to '{}'".format(workload_id, thread_ids))
            self.__cgroup_manager.set_cpuset(workload_id, thread_ids)

        return len(updates) > 0

    def get_workloads(self):
        return list(self.__workloads.values())

    def get_workload_map_copy(self):
        return copy.deepcopy(self.__workloads)

    def get_isolated_workload_ids(self):
        return self.__cgroup_manager.get_isolated_workload_ids()

    def is_isolated(self, workload_id):
        noop_allocators = [NoopCpuAllocator, NoopResetCpuAllocator]
        for allocator in noop_allocators:
            if isinstance(self.__cpu_allocator, allocator):
                return True

        return workload_id in self.get_isolated_workload_ids()

    def get_cpu(self):
        return self.__cpu

    def get_cpu_copy(self):
        return copy.deepcopy(self.__cpu)

    def get_added_count(self):
        return self.__added_count

    def get_removed_count(self):
        return self.__removed_count

    def get_rebalanced_count(self):
        return self.__removed_count

    def get_success_count(self):
        return self.get_added_count() + \
               self.get_removed_count() + \
               self.get_rebalanced_count()

    def get_error_count(self):
        return self.__error_count

    def get_allocator_call_duration_sum_secs(self):
        return self.__allocator_call_duration_sum_secs

    def get_allocator_name(self):
        if isinstance(self.__cpu_allocator, FallbackCpuAllocator):
            return self.__cpu_allocator.get_primary_allocator().__class__.__name__

        return self.__cpu_allocator.__class__.__name__

    def set_registry(self, registry):
        self.__reg = registry
        self.__cpu_allocator.set_registry(registry)
        self.__cgroup_manager.set_registry(registry)

    def __report_cpu_state(self, old_cpu, new_cpu):
        log.info(visualize_cpu_comparison(old_cpu, new_cpu))
        Thread(target=report_cpu, args=[new_cpu, self.get_workload_map_copy().values()]).start()

    def report_metrics(self, tags):
        cpu = self.get_cpu_copy()
        workload_map = self.get_workload_map_copy()
        Thread(target=report_cpu, args=[cpu, workload_map.values()]).start()

        self.__reg.gauge(RUNNING, tags).set(1)

        self.__reg.gauge(ADDED_KEY, tags).set(self.get_added_count())
        self.__reg.gauge(REMOVED_KEY, tags).set(self.get_removed_count())
        self.__reg.gauge(REBALANCED_KEY, tags).set(self.get_rebalanced_count())
        self.__reg.gauge(SUCCEEDED_KEY, tags).set(self.get_success_count())
        self.__reg.gauge(FAILED_KEY, tags).set(self.get_error_count())
        self.__reg.gauge(WORKLOAD_COUNT_KEY, tags).set(len(self.get_workloads()))
        self.__reg.gauge(ADDED_TO_FULL_CPU_ERROR_KEY, tags).set(self.__added_to_full_cpu_count)

        self.__reg.gauge(ALLOCATOR_CALL_DURATION, tags).set(self.get_allocator_call_duration_sum_secs())

        cross_package_violation_count = len(get_cross_package_violations(cpu))
        shared_core_violation_count = len(get_shared_core_violations(cpu))
        self.__reg.gauge(PACKAGE_VIOLATIONS_KEY, tags).set(cross_package_violation_count)
        self.__reg.gauge(CORE_VIOLATIONS_KEY, tags).set(shared_core_violation_count)

        # Allocation / Request sizes
        self.__reg.gauge(ALLOCATED_SIZE_KEY, tags).set(get_allocated_size(cpu))
        self.__reg.gauge(UNALLOCATED_SIZE_KEY, tags).set(get_unallocated_size(cpu))
        self.__reg.gauge(STATIC_ALLOCATED_SIZE_KEY, tags).set(get_static_allocated_size(cpu, workload_map))
        self.__reg.gauge(BURST_ALLOCATED_SIZE_KEY, tags).set(get_burst_allocated_size(cpu, workload_map))
        self.__reg.gauge(BURST_REQUESTED_SIZE_KEY, tags).set(get_burst_request_size(list(workload_map.values())))
        self.__reg.gauge(OVERSUBSCRIBED_THREADS_KEY, tags).set(get_oversubscribed_thread_count(cpu, workload_map))

        # Have the sub-components report metrics
        self.__cpu_allocator.report_metrics(tags)
        self.__cgroup_manager.report_metrics(tags)



import os
import time
from typing import List

from titus_isolate import log
from titus_isolate.monitor.cpu_usage import CpuUsage

ROOT_CGROUP_PATH = "/sys/fs/cgroup"
TITUS_INITS_PATH = "/var/lib/titus-inits"
TITUS_ENVIRONMENTS_PATH = "/var/lib/titus-environments"

CPUSET = "cpuset"
CPU_CPUACCT = "cpu,cpuacct"


def __get_info_path(container_name):
    return "{}/{}/cgroup".format(TITUS_INITS_PATH, container_name)


def __get_json_path(container_name):
    return "{}/{}.json".format(TITUS_ENVIRONMENTS_PATH, container_name)


def __noop():
    pass


def _wait_for_file_to_exist(path, timeout, check_func=__noop):
    start_time = time.time()

    log.info("Waiting '{}' seconds for file to exist: '{}'".format(timeout, path))
    while not os.path.exists(path):
        time.sleep(0.1)

        check_func()

        elapsed_time = time.time() - start_time
        if elapsed_time > timeout:
            raise TimeoutError(
                "Expected file '{}' was not created in '{}' seconds.".format(path, timeout))


def wait_for_files(container_name, cgroup_timeout, json_timeout):
    info_file_path = __get_info_path(container_name)
    json_file_path = __get_json_path(container_name)

    def __raise_if_json_file_gone():
        if not os.path.exists(json_file_path):
            raise RuntimeError("JSON file: '{}' disappeared, meaning the task exited.".format(json_file_path))

    _wait_for_file_to_exist(json_file_path, json_timeout)
    _wait_for_file_to_exist(info_file_path, cgroup_timeout, __raise_if_json_file_gone)


def _get_cgroup_path_from_list(cgroups_list, cgroup_name):
    for row in cgroups_list:
        r = row.split(":")
        name = r[1]
        path = r[2]

        if name == cgroup_name:
            return path.strip()

    return None


def get_cgroup_path_from_file(file_path, cgroup_name):
    log.debug("Reading '{}' path from file: '{}'".format(cgroup_name, file_path))
    with open(file_path, "r") as myfile:
        data = myfile.readlines()
        return _get_cgroup_path_from_list(data, cgroup_name)


def get_cpuset_path(container_name):
    file_path = __get_info_path(container_name)
    cgroup_path = get_cgroup_path_from_file(file_path, CPUSET)
    return "{}/cpuset{}/cpuset.cpus".format(ROOT_CGROUP_PATH, cgroup_path)


def get_quota_path(container_name):
    file_path = __get_info_path(container_name)
    cgroup_path = get_cgroup_path_from_file(file_path, CPU_CPUACCT)
    return "{}/cpu,cpuacct{}/cpu.cfs_quota_us".format(ROOT_CGROUP_PATH, cgroup_path)


def get_usage_all_path(container_name):
    file_path = __get_info_path(container_name)
    cgroup_path = get_cgroup_path_from_file(file_path, CPU_CPUACCT)
    return "{}/cpu,cpuacct{}/cpuacct.usage_all".format(ROOT_CGROUP_PATH, cgroup_path)


def set_cpuset(container_name, threads_str):
    path = get_cpuset_path(container_name)

    log.info("Writing '{}' to path '{}'".format(threads_str, path))
    with open(path, 'w') as f:
        f.write(threads_str)


def get_cpuset(container_name):
    path = get_cpuset_path(container_name)

    log.debug("Reading path '{}'".format(path))
    with open(path, 'r') as f:
        return f.read()


def set_quota(container_name, value):
    path = get_quota_path(container_name)

    log.info("Writing '{}' to path '{}'".format(value, path))
    with open(path, 'w') as f:
        f.write(str(value))


def get_quota(container_name):
    path = get_quota_path(container_name)

    log.info("Reading from path '{}'".format(path))
    with open(path, 'r') as f:
        return int(f.readline().strip())


def parse_cpuacct_usage_all(text) -> List[CpuUsage]:
    # Text looks like this:
    #
    #     cpu user system
    #     0 0 0
    #     1 0 0
    #     2 73169258935876 0
    #     3 0 0
    #     ...
    #
    # so we skip the first line and return a list of lists

    #raw_rows = [line.split() for line in text.splitlines()[1:]]

    usage_rows = []
    for ind, row in enumerate(text.splitlines()):
        if ind == 0:
            continue
        cpu, user, system = row.split()
        usage_rows.append(CpuUsage(int(cpu), int(user), int(system)))

    return usage_rows


def parse_cpuset(cpuset_str: str) -> list:
    ranges = list(x.split("-") for x in cpuset_str.split(","))
    if len(ranges) == 0:
        return []
    else:
        return list([i for r in ranges for i in range(int(r[0]), int(r[-1]) + 1)])



import collections
from threading import Lock

import schedule

from titus_isolate import log
from titus_isolate.config.constants import DEFAULT_SAMPLE_FREQUENCY_SEC
from titus_isolate.event.constants import BURST, STATIC
from titus_isolate.metrics.constants import BURST_POOL_USAGE_KEY, STATIC_POOL_USAGE_KEY
from titus_isolate.metrics.metrics_reporter import MetricsReporter
from titus_isolate.monitor.cgroup_metrics_provider import CgroupMetricsProvider
from titus_isolate.monitor.cpu_usage_provider import CpuUsageProvider
from titus_isolate.monitor.workload_perf_mon import WorkloadPerformanceMonitor
from titus_isolate.utils import get_workload_manager


class WorkloadMonitorManager(CpuUsageProvider, MetricsReporter):

    def __init__(self, sample_interval=DEFAULT_SAMPLE_FREQUENCY_SEC):
        self.__sample_interval = sample_interval
        self.__lock = Lock()
        self.__monitors = {}
        self.__registry = None

        schedule.every(sample_interval).seconds.do(self.__sample)

    def get_cpu_usage(self, seconds: int, agg_granularity_secs: int) -> dict:
        cpu_usage = {}
        for workload_id, monitor in self.get_monitors().items():
            cpu_usage[workload_id] = monitor.get_normalized_cpu_usage_last_seconds(seconds, agg_granularity_secs)

        return cpu_usage

    def set_registry(self, registry):
        self.__registry = registry

    def report_metrics(self, tags):
        if self.__registry is None:
            log.debug("Not reporting metrics because there's no registry available yet.")
            return

        wm = get_workload_manager()
        if wm is None:
            log.debug("Not reporting metrics because there's no workload manager available yet.")
            return

        usage = self.get_cpu_usage(60, 60)
        static_pool_cpu_usage = self.__get_pool_usage(STATIC, usage)
        burst_pool_cpu_usage = self.__get_pool_usage(BURST, usage)

        self.__registry.gauge(STATIC_POOL_USAGE_KEY, tags).set(static_pool_cpu_usage)
        self.__registry.gauge(BURST_POOL_USAGE_KEY, tags).set(burst_pool_cpu_usage)

    @staticmethod
    def __get_pool_usage(workload_type, usage):
        wm = get_workload_manager()
        if wm is None:
            log.debug("Not reporting metrics because there's no workload manager available yet.")
            return

        workload_map = wm.get_workload_map_copy()

        pool_cpu_usage = 0.0
        for w_id, usage in usage.items():
            if w_id not in workload_map:
                continue

            workload = workload_map[w_id]
            if workload.get_type() == workload_type:
                pool_cpu_usage += usage[len(usage) - 1]

        return pool_cpu_usage

    def get_monitors(self):
        return self.__monitors

    def to_dict(self):
        with self.__lock:
            monitors_dict = {}
            for workload_id, monitor in sorted(self.get_monitors().items()):
                buffers = collections.OrderedDict()
                _, timestamps, mon_buffers = monitor.get_buffers()
                for cpu_ind, buff in enumerate(mon_buffers):
                    buffers[str(cpu_ind)] = [str(i) for i in buff]
                buffers['timestamps'] = timestamps
                monitors_dict[workload_id] = buffers

            return monitors_dict

    def __sample(self):
        try:
            self.__update_monitors()
            self.__sample_monitors()
        except:
            log.exception("Failed to sample performance monitors.")

    @staticmethod
    def __get_workloads():
        wm = get_workload_manager()
        if wm is None:
            log.debug("Workload manager not yet present.")
            return []

        return wm.get_workloads()

    def __update_monitors(self):
        workloads = self.__get_workloads()

        with self.__lock:
            # Remove monitors for workloads which are no longer managed
            workload_ids = [w.get_id() for w in workloads]
            for monitored_id in list(self.__monitors.keys()):
                if monitored_id not in workload_ids:
                    self.__monitors.pop(monitored_id, None)

            # Add monitors for new workloads
            for workload in workloads:
                if workload.get_id() not in self.__monitors:
                    self.__monitors[workload.get_id()] = \
                        WorkloadPerformanceMonitor(CgroupMetricsProvider(workload), self.__sample_interval)

    def __sample_monitors(self):
        with self.__lock:
            for workload_id, monitor in self.__monitors.items():
                try:
                    monitor.sample()
                except:
                    log.exception("Failed to sample performance of workload: '{}'".format(workload_id))



import os

import numpy as np

from titus_isolate import log
from titus_isolate.model.workload import Workload

from titus_optimize.data import Query, build_ts_features
from titus_optimize.predictors import PredictorWithFilter


class PredEnvironment:
    def __init__(self, region, nflx_env, hour_of_day):
        self.region = region
        self.nflx_env = nflx_env
        self.hour_of_day = hour_of_day


class CpuUsagePredictor:

    def __init__(self, model_path, use_whitelist=True):
        pf = CpuUsagePredictor.__load_model_from_file(model_path)
        self.__use_whitelist = use_whitelist
        self.__model = pf
    
    def predict(self, workload: Workload, cpu_usage_last_hour: np.array, pred_env: PredEnvironment) -> float:
        image = workload.get_image()
        tokens = image.split('@')
        valid_digest = False
        if cpu_usage_last_hour is None:
            cpu_usage_last_hour = np.full((60,), np.nan, dtype=np.float32)
        if len(tokens) == 2 and tokens[-1].startswith("sha256:"):
            valid_digest = True
            entry_point = workload.get_entrypoint()
            if entry_point is None:
                entry_point = ""
            filter_key = "%s@%s" % (tokens[-1], entry_point)
        if self.__use_whitelist and valid_digest and (filter_key not in self.__model.filter):
            # not in whitelist, predict without context features
            q = Query(
                None, # user
                None, # app_name
                workload.get_thread_count(),
                None, # ram_requested
                None, # disk_requested
                None, # network_requested
                None, # job_type
                None, # region
                None, # env
                None, # hour of day
                build_ts_features(cpu_usage_last_hour)
            )
        else:
            q = Query(
                workload.get_owner_email(),
                workload.get_app_name(),
                workload.get_thread_count(),
                workload.get_mem(),
                workload.get_disk(),
                workload.get_network(),
                workload.get_job_type().lower(),
                pred_env.region,
                pred_env.nflx_env,
                pred_env.hour_of_day,
                build_ts_features(cpu_usage_last_hour)
            )

        return min(self.__model.ml_model.predict_single(q), workload.get_thread_count())
    
    def get_model(self):
        return self.__model

    @staticmethod
    def __load_model_from_file(path):
        if not os.path.isfile(path):
            raise Exception("Could not find a model at `%s`" % (path,))
        model =  PredictorWithFilter.load(open(path, "rb").read())
        log.info("Loaded CPU usage predictor. Meta-data: %s. Whitelist size: %s" % (model.meta_data, len(model.filter)))
        return model



from threading import Lock

import schedule

from titus_isolate.allocate.utils import download_latest_cpu_model, get_cpu_model_file_path
from titus_isolate.predict.cpu_usage_predictor import CpuUsagePredictor


class CpuUsagePredictorManager:

    def __init__(self):
        self.__lock = Lock()
        self.__predictor = None

        self.__update_predictor()
        schedule.every(1).hour.do(self.__update_predictor)

    def __update_predictor(self):
        download_latest_cpu_model()
        with self.__lock:
            self.__predictor = CpuUsagePredictor(get_cpu_model_file_path())

    def get_predictor(self):
        return self.__predictor



from titus_isolate import log
from titus_isolate.allocate.cpu_allocator import CpuAllocator
from titus_isolate.event.constants import STATIC
from titus_isolate.model.processor.cpu import Cpu
from titus_isolate.model.processor.utils import get_emptiest_core, is_cpu_full
from titus_isolate.model.utils import get_burst_workloads, release_all_threads, update_burst_workloads, rebalance
from titus_isolate.model.workload import Workload
from titus_isolate.monitor.empty_free_thread_provider import EmptyFreeThreadProvider
from titus_isolate.monitor.free_thread_provider import FreeThreadProvider


class GreedyCpuAllocator(CpuAllocator):

    def __init__(self, free_thread_provider: FreeThreadProvider = EmptyFreeThreadProvider()):
        self.__free_thread_provider = free_thread_provider

    def assign_threads(self, cpu, workload_id, workloads):
        burst_workloads = get_burst_workloads(workloads.values())
        release_all_threads(cpu, burst_workloads)
        if workloads[workload_id].get_type() == STATIC:
            self.__assign_threads(cpu, workloads[workload_id])
        update_burst_workloads(cpu, burst_workloads, self.__free_thread_provider)
        return cpu

    def free_threads(self, cpu, workload_id, workloads):
        burst_workloads = get_burst_workloads(workloads.values())
        release_all_threads(cpu, burst_workloads)
        for t in cpu.get_threads():
            if workload_id in t.get_workload_ids():
                t.free(workload_id)

        burst_workloads = [w for w in burst_workloads if w.get_id() != workload_id]
        update_burst_workloads(cpu, burst_workloads, self.__free_thread_provider)

        return cpu

    def rebalance(self, cpu: Cpu, workloads: dict) -> Cpu:
        return rebalance(cpu, workloads, self.__free_thread_provider)

    def __assign_threads(self, cpu, workload):
        thread_count = workload.get_thread_count()
        claimed_threads = []

        if thread_count == 0:
            return claimed_threads

        if is_cpu_full(cpu):
            raise ValueError("Failed to add workload: '{}', cpu is full: {}".format(workload.get_id(), cpu))

        package = cpu.get_emptiest_package()

        while thread_count > 0 and len(package.get_empty_threads()) > 0:
            core = get_emptiest_core(package)
            empty_threads = core.get_empty_threads()[:thread_count]

            for empty_thread in empty_threads:
                log.debug("Claiming package:core:thread '{}:{}:{}' for workload '{}'".format(
                    package.get_id(), core.get_id(), empty_thread.get_id(), workload.get_id()))
                empty_thread.claim(workload.get_id())
                claimed_threads.append(empty_thread)
                thread_count -= 1

        return claimed_threads + self.__assign_threads(
            cpu,
            Workload(
                identifier=workload.get_id(),
                thread_count=thread_count,
                mem=workload.get_mem(),
                disk=workload.get_disk(),
                network=workload.get_network(),
                app_name=workload.get_app_name(),
                owner_email=workload.get_owner_email(),
                image=workload.get_image(),
                command=workload.get_command(),
                entrypoint=workload.get_entrypoint(),
                job_type=workload.get_job_type(),
                workload_type=workload.get_type()))

    def set_registry(self, registry):
        pass

    def report_metrics(self, tags):
        pass



from titus_isolate.allocate.cpu_allocator import CpuAllocator
from titus_isolate.config.constants import DEFAULT_MAX_SOLVER_RUNTIME, MAX_SOLVER_RUNTIME
from titus_isolate.utils import get_config_manager
from titus_optimize.compute import IP_SOLUTION_TIME_BOUND, optimize_ip

from titus_isolate.event.constants import STATIC
from titus_isolate.metrics.constants import IP_ALLOCATOR_TIMEBOUND_COUNT
from titus_isolate.model.processor.cpu import Cpu
from titus_isolate.model.processor.utils import is_cpu_full
from titus_isolate.model.utils import get_sorted_workloads, get_burst_workloads, release_all_threads, \
    update_burst_workloads, rebalance
from titus_isolate.monitor.empty_free_thread_provider import EmptyFreeThreadProvider


class IntegerProgramCpuAllocator(CpuAllocator):

    def __init__(self, free_thread_provider=EmptyFreeThreadProvider()):

        self.__reg = None
        self.__cache = {}
        self.__time_bound_call_count = 0

        self.__solver_max_runtime_secs = get_config_manager().get(MAX_SOLVER_RUNTIME, DEFAULT_MAX_SOLVER_RUNTIME)
        self.__free_thread_provider = free_thread_provider

    def assign_threads(self, cpu, workload_id, workloads):
        burst_workloads = get_burst_workloads(workloads.values())
        release_all_threads(cpu, burst_workloads)
        if workloads[workload_id].get_type() == STATIC:
            self.__assign_threads(cpu, workload_id, workloads)
        update_burst_workloads(cpu, burst_workloads, self.__free_thread_provider)
        return cpu

    def free_threads(self, cpu, workload_id, workloads):
        burst_workloads = get_burst_workloads(workloads.values())
        release_all_threads(cpu, burst_workloads)
        if workloads[workload_id].get_type() == STATIC:
            self.__free_threads(cpu, workload_id, workloads)
        burst_workloads = [w for w in burst_workloads if w.get_id() != workload_id]
        update_burst_workloads(cpu, burst_workloads, self.__free_thread_provider)
        return cpu

    def rebalance(self, cpu: Cpu, workloads: dict) -> Cpu:
        return rebalance(cpu, workloads, self.__free_thread_provider)

    def __assign_threads(self, cpu, workload_id, workloads):
        """
        Use the integer-program solver to find the optimal static placement
        when adding the given workload on the cpu.

        `workload_insertion_times` is a workload_id --> integer dict
        indicating unix timestamps at which workloads currently running on the cpu
        have been placed.
        """

        if is_cpu_full(cpu):
            raise ValueError("CPU is full, failed to add workload: '{}'".format(workload_id))

        n_compute_units = len(cpu.get_threads())
        curr_ids_per_workload = cpu.get_workload_ids_to_thread_ids()

        from titus_isolate.model.utils import get_sorted_workloads
        ordered_workload_ids = [w.get_id() for w in get_sorted_workloads(workloads.values())]
        tid_2order = {i: t.get_id() for i, t in enumerate(cpu.get_threads())}

        curr_placement_vectors = []
        for wid in ordered_workload_ids:
            cids = curr_ids_per_workload[wid]
            v = [1 if tid_2order[i] in cids else 0 for i in range(n_compute_units)]
            curr_placement_vectors.append(v)
        if len(curr_placement_vectors) == 0:
            curr_placement_vectors = None
            requested_cus = []
        else:
            requested_cus = [sum(v) for v in curr_placement_vectors]

        workload = workloads[workload_id]
        requested_cus += [workload.get_thread_count()]

        new_placement_vectors = self.__compute_new_placement(cpu, curr_placement_vectors, requested_cus)

        ordered_workload_ids.append(workload.get_id())

        thread_id2workload_id = {}
        for w_ind, v in enumerate(new_placement_vectors):
            for i, e in enumerate(v):
                if e == 1:
                    thread_id2workload_id[tid_2order[i]] = ordered_workload_ids[w_ind]

        self.__assign_new_mapping(cpu, thread_id2workload_id)

        return cpu

    def __free_threads(self, cpu, workload_id, workloads):
        """
        Use the integerprogram solver to find the optimal static placement
        after removing the given workload from the cpu.
        """

        n_compute_units = len(cpu.get_threads())

        curr_ids_per_workload = cpu.get_workload_ids_to_thread_ids()

        if workload_id not in curr_ids_per_workload:
            raise Exception("workload_id=`%s` is not placed on the instance. Cannot free it." % (workload_id,))

        ordered_workload_ids = [w.get_id() for w in get_sorted_workloads(workloads.values())]
        tid_2order = {i: t.get_id() for i, t in enumerate(cpu.get_threads())}

        curr_placement_vectors = []
        for wid in ordered_workload_ids:
            cids = curr_ids_per_workload[wid]
            v = [1 if tid_2order[i] in cids else 0 for i in range(n_compute_units)]
            curr_placement_vectors.append(v)
        if len(curr_placement_vectors) == 0:
            raise Exception("Cannot free a workload from an empty CPU")

        requested_cus = [len(curr_ids_per_workload[wid]) if wid != workload_id else 0 for wid in ordered_workload_ids]

        new_placement_vectors = self.__compute_new_placement(cpu, curr_placement_vectors, requested_cus)

        thread_id2workload_id = {}
        for w_ind, v in enumerate(new_placement_vectors):
            if ordered_workload_ids[w_ind] == workload_id:
                continue
            for i, e in enumerate(v):
                if e == 1:
                    thread_id2workload_id[tid_2order[i]] = ordered_workload_ids[w_ind]

        self.__assign_new_mapping(cpu, thread_id2workload_id)
        return cpu

    @staticmethod
    def __assign_new_mapping(cpu, thread_id2workload_id):
        cpu.clear()
        for t in cpu.get_threads():
            wid = thread_id2workload_id.get(t.get_id(), None)
            if wid is not None:
                t.claim(wid)

    def __compute_new_placement(self, cpu, current_placement, requested_units):
        key_req = '-'.join([str(e) for e in requested_units])
        cache_key = key_req
        if current_placement is not None:
            cache_key += '&' + '%'.join(['|'.join([str(e) for e in v]) for v in current_placement])

        cache_val = self.__cache.get(cache_key, None)
        if cache_val is None:
            placement, status = optimize_ip(
                requested_units,
                len(cpu.get_threads()),
                len(cpu.get_packages()),
                current_placement,
                verbose=False,
                max_runtime_secs=self.__solver_max_runtime_secs)
            self.__cache[cache_key] = (placement, status)
        else:
            placement, status = cache_val[0], cache_val[1]

        if status == IP_SOLUTION_TIME_BOUND:
            self.__time_bound_call_count += 1

        return placement

    def set_solver_max_runtime_secs(self, val):
        self.__solver_max_runtime_secs = val

    def set_registry(self, registry):
        self.__reg = registry

    def report_metrics(self, tags):
        self.__reg.gauge(IP_ALLOCATOR_TIMEBOUND_COUNT, tags).set(self.__time_bound_call_count)




import abc

from titus_isolate.metrics.metrics_reporter import MetricsReporter
from titus_isolate.model.processor.cpu import Cpu


class CpuAllocator(abc.ABC, MetricsReporter):

    @abc.abstractmethod
    def assign_threads(self, cpu: Cpu, workload_id: str, workloads: dict) -> Cpu:
        """
        Implementations of this method should claim threads for a workload on a given CPU.

        :param cpu: An object indicating the state of the CPU before workload assignment
        :param workload_id: The id of the workload being assigned.
        :param workloads: A map of all relevant workloads including the workload to be assigned.
        The keys are workload ids, the objects are Workload objects.
        """
        pass

    @abc.abstractmethod
    def free_threads(self, cpu: Cpu, workload_id: str, workloads: dict) -> Cpu:
        """
        Implementations of this method should free threads claimed by a workload on a given CPU.

        :param cpu: An object indicating the state of the CPU before freeing threads
        :param workload_id: The id of the workload being removed.
        :param workloads: A map of all relevant workloads including the workload to be removed.
        The keys are workload ids, the values are Workload objects.
        """
        pass

    @abc.abstractmethod
    def rebalance(self, cpu: Cpu, workloads: dict) -> Cpu:
        """
        This method will be called periodically to provide an opportunity to move already running
        workloads according to whatever policy the allocator deems appropriate.  Examples policies
        may choose to oversubscribe resources or improve isolation.

        :param cpu: An object indicating the state of the CPU before freeing threads
        :param workloads: A map of all relevant workloads including the workload to be removed.
        The keys are workload ids, the values are Workload objects.
        """
        pass

    def str(self):
        return self.__class__.__name__



import json

ACTION = "Action"
ACTOR = "Actor"
ATTRIBUTES = "Attributes"
ID = "ID"
LOWERCASE_ID = "id"
NAME = "name"
REPO_DIGESTS = 'RepoDigests'
TIME = "time"
TYPE = "Type"

CONTAINER = "container"
CREATE = "create"
DIE = "die"
REBALANCE = "rebalance"
RECONCILE = "reconcile"

APP_NAME_LABEL_KEY = "com.netflix.titus.appName"
CPU_LABEL_KEY = "com.netflix.titus.cpu"
MEM_LABEL_KEY = "com.netflix.titus.mem"
DISK_LABEL_KEY = "com.netflix.titus.disk"
NETWORK_LABEL_KEY = "com.netflix.titus.network"
JOB_TYPE_LABEL_KEY = "com.netflix.titus.job.type"
WORKLOAD_TYPE_LABEL_KEY = "com.netflix.titus.workload.type"
OWNER_EMAIL_LABEL_KEY = "com.netflix.titus.owner.email"
COMMAND_LABEL_KEY = "com.netflix.titus.command"
ENTRYPOINT_LABEL_KEY = "com.netflix.titus.entrypoint"
IMAGE_LABEL_KEY = "image"

REQUIRED_LABELS = [
    CPU_LABEL_KEY,
    MEM_LABEL_KEY,
    DISK_LABEL_KEY,
    NETWORK_LABEL_KEY,
    WORKLOAD_TYPE_LABEL_KEY]

STATIC = "static"
BURST = "burst"
WORKLOAD_TYPES = [STATIC, BURST]

REBALANCE_EVENT = json.dumps({ACTION: REBALANCE}).encode("utf-8")
RECONCILE_EVENT = json.dumps({ACTION: RECONCILE}).encode("utf-8")



import os

import schedule
from spectator import GlobalRegistry

from titus_isolate import log
from titus_isolate.isolate.utils import get_allocator
from titus_isolate.utils import get_config_manager

registry = GlobalRegistry


class MetricsManager:

    def __init__(self, reporters, reg=registry, report_interval=60):
        self.__reporters = reporters
        self.__reg = reg

        for reporter in self.__reporters:
            reporter.set_registry(self.__reg)

        schedule.every(report_interval).seconds.do(self.__report_metrics)

    def __report_metrics(self):
        try:
            tags = self.__get_tags()

            for reporter in self.__reporters:
                reporter.report_metrics(tags)
        except:
            log.exception("Failed to report metrics.")

    @staticmethod
    def __get_tags():
        ec2_instance_id = 'EC2_INSTANCE_ID'

        tags = {}
        if ec2_instance_id in os.environ:
            tags["node"] = os.environ[ec2_instance_id]

        allocator_name = get_allocator(get_config_manager()).get_primary_allocator().__class__.__name__
        tags["cpu_allocator"] = allocator_name

        return tags



ADDED_KEY = 'titus-isolate.added'
REMOVED_KEY = 'titus-isolate.removed'
REBALANCED_KEY = 'titus-isolate.rebalanced'
SUCCEEDED_KEY = 'titus-isolate.succeeded'
FAILED_KEY = 'titus-isolate.failed'
ALLOCATOR_CALL_DURATION = 'titus-isolate.allocatorCallDurationSecs'
QUEUE_DEPTH_KEY = 'titus-isolate.queueDepth'
WORKLOAD_COUNT_KEY = 'titus-isolate.workloadCount'
EVENT_SUCCEEDED_KEY = 'titus-isolate.eventSucceeded'
EVENT_FAILED_KEY = 'titus-isolate.eventFailed'
EVENT_PROCESSED_KEY = 'titus-isolate.eventProcessed'
ADDED_TO_FULL_CPU_ERROR_KEY = 'titus-isolate.addedToFullCpu'

IP_ALLOCATOR_TIMEBOUND_COUNT = 'titus-isolate.ipAllocatorTimeBoundSolutionCount'
FORECAST_REBALANCE_FAILURE_COUNT = 'titus-isolate.forecastRebalanceFailureCount'

WRITE_CPUSET_SUCCEEDED_KEY = 'titus-isolate.writeCpusetSucceeded'
WRITE_CPUSET_FAILED_KEY = 'titus-isolate.writeCpusetFailed'
ISOLATED_WORKLOAD_COUNT = 'titus-isolate.isolatedWorkloadCount'

PACKAGE_VIOLATIONS_KEY = 'titus-isolate.crossPackageViolations'
CORE_VIOLATIONS_KEY = 'titus-isolate.sharedCoreViolations'

ALLOCATED_SIZE_KEY = 'titus-isolate.allocatedSize'
UNALLOCATED_SIZE_KEY = 'titus-isolate.unallocatedSize'
STATIC_ALLOCATED_SIZE_KEY = 'titus-isolate.staticAllocatedSize'
BURST_ALLOCATED_SIZE_KEY = 'titus-isolate.burstAllocatedSize'
BURST_REQUESTED_SIZE_KEY = 'titus-isolate.burstRequestedSize'
OVERSUBSCRIBED_THREADS_KEY = 'titus-isolate.oversubscribedThreads'

FALLBACK_ALLOCATOR_COUNT = 'titus-isolate.fallbackCount'
FALLBACK_ASSIGN_COUNT = 'titus-isolate.assignThreadsFallbackCount'
FALLBACK_FREE_COUNT = 'titus-isolate.freeThreadsFallbackCount'
FALLBACK_REBALANCE_COUNT = 'titus-isolate.rebalanceFallbackCount'

STATIC_POOL_USAGE_KEY = 'titus-isolate.staticPoolUsage'
BURST_POOL_USAGE_KEY = 'titus-isolate.burstPoolUsage'

RUNNING = 'titus-isolate.running'



from titus_isolate import log
from titus_isolate.model.processor.cpu import Cpu
from titus_isolate.model.processor.thread import Thread
from titus_isolate.monitor.free_thread_provider import FreeThreadProvider
from titus_isolate.utils import get_workload_monitor_manager


class ThresholdFreeThreadProvider(FreeThreadProvider):

    def __init__(
            self,
            total_threshold: float,
            total_duration_sec: int,
            per_workload_threshold: float,
            per_workload_duration_sec: int):

        self.__total_threshold = total_threshold
        self.__total_duration_sec = total_duration_sec

        self.__per_workload_threshold = per_workload_threshold
        self.__per_workload_duration_sec = per_workload_duration_sec

        log.debug("ThresholdFreeThreadProvider created with total_threshold: '{}', total_duration_sec: '{}', per_workload_threshold: '{}', per_workload_duration_sec: '{}'".format(
           self.__total_threshold, self.__total_duration_sec, self.__per_workload_threshold, self.__per_workload_duration_sec))

    def get_free_threads(self, cpu: Cpu) -> list:
        wmm = get_workload_monitor_manager()
        if wmm is None:
            log.debug("Workload monitor manager is not yet set.")
            return []

        total_usage = wmm.get_cpu_usage(self.__total_duration_sec, self.__total_duration_sec)
        workload_usage = wmm.get_cpu_usage(self.__per_workload_duration_sec, self.__per_workload_duration_sec)

        # A thread is free if no workload has claimed it or its combined usage is below the threshold
        free_threads = [t for t in cpu.get_threads() if self.__is_free(t, workload_usage, total_usage)]
        free_thread_ids = [t.get_id() for t in free_threads]
        log.debug("Found free threads: {}".format(free_thread_ids))

        return free_threads

    @staticmethod
    def __is_reporting_metrics(workload_id, usage_dicts: list) -> bool:
        for d in usage_dicts:
            if workload_id not in d:
                return False

            if d[workload_id] is None:
                return False

        return True

    def __is_free(self, thread: Thread, workload_usage: dict, total_usage: dict) -> bool:
        log.debug("Determining if thread: '{}' is free.".format(thread.get_id()))
        workload_ids = thread.get_workload_ids()

        # If no workload is using the thread, it's free.
        if len(workload_ids) == 0:
            return True

        # Check individual workloads to see if any is using the thread.
        for w_id in workload_ids:

            # A workload probably just started so the thread isn't free
            if not self.__is_reporting_metrics(w_id, [workload_usage, total_usage]):
                log.debug("Thread '{}' is not free because workload: '{}' is does not have usage metrics.".format(
                    thread.get_id(), w_id))
                return False

            usage = workload_usage[w_id][str(thread.get_id())]

            if usage > self.__per_workload_threshold:
                log.debug("Thread '{}' is not free because the cpu usage: '{}' of workload: '{}' exceeds the per-workload usage threshold: '{}'.".format(
                    thread.get_id(), usage, w_id, self.__per_workload_threshold))
                return False

        # If the cpu's total usage exceeds a threshold it's not free.
        aggregate_usage = self.__get_aggregate_usage(thread, total_usage)
        if aggregate_usage > self.__total_threshold:
            log.debug("Thread '{}' is not free because it's usage: '{}' exceeds the total thread usage threshold: '{}'."
                      .format(thread.get_id(), aggregate_usage, self.__total_threshold))
            return False

        log.debug("Thread '{}' is free.".format(thread.get_id()))
        return True

    def __get_aggregate_usage(self, thread: Thread, total_usage: dict) -> float:
        agg_usage = 0.0
        for w_id in thread.get_workload_ids():
            agg_usage += total_usage[w_id][str(thread.get_id())]

        return agg_usage



import json
from queue import Queue, Empty
from threading import Thread, Lock

import schedule

from titus_isolate import log
from titus_isolate.config.constants import REBALANCE_FREQUENCY_KEY, DEFAULT_REBALANCE_FREQUENCY, \
    RECONCILE_FREQUENCY_KEY, DEFAULT_RECONCILE_FREQUENCY
from titus_isolate.event.constants import REBALANCE_EVENT, RECONCILE_EVENT
from titus_isolate.metrics.constants import QUEUE_DEPTH_KEY, EVENT_SUCCEEDED_KEY, EVENT_FAILED_KEY, EVENT_PROCESSED_KEY
from titus_isolate.metrics.metrics_reporter import MetricsReporter
from titus_isolate.utils import get_config_manager

DEFAULT_EVENT_TIMEOUT_SECS = 60


class EventManager(MetricsReporter):

    def __init__(self, event_iterable, event_handlers, event_timeout=DEFAULT_EVENT_TIMEOUT_SECS):
        self.__reg = None
        self.__stopped = False
        self.__q = Queue()

        self.__events = event_iterable
        self.__event_handlers = event_handlers
        self.__event_timeout = event_timeout

        self.__success_event_count = 0
        self.__error_event_count = 0
        self.__processed_event_count = 0

        self.__started = False
        self.__started_lock = Lock()

        self.__processing_thread = Thread(target=self.__process_events)
        self.__pulling_thread = Thread(target=self.__pull_events)

        config_manager = get_config_manager()

        rebalance_frequency = int(config_manager.get(REBALANCE_FREQUENCY_KEY, DEFAULT_REBALANCE_FREQUENCY))
        schedule.every(rebalance_frequency).seconds.do(self.__rebalance)

        reconcile_frequency = int(config_manager.get(RECONCILE_FREQUENCY_KEY, DEFAULT_RECONCILE_FREQUENCY))
        schedule.every(reconcile_frequency).seconds.do(self.__reconcile)

    def join(self):
        self.__pulling_thread.join()
        self.__processing_thread.join()

    def stop_processing_events(self):
        self.__stopped = True
        self.__events.close()
        self.join()

    def start_processing_events(self):
        with self.__started_lock:
            if self.__started:
                return

            self.__processing_thread.start()
            self.__pulling_thread.start()
            self.__started = True

    def get_success_count(self):
        return self.__success_event_count

    def get_error_count(self):
        return self.__error_event_count

    def get_processed_count(self):
        return self.__processed_event_count

    def get_queue_depth(self):
        return self.__q.qsize()

    def __rebalance(self):
        self.__q.put(REBALANCE_EVENT)

    def __reconcile(self):
        if self.__q.empty():
            self.__q.put(RECONCILE_EVENT)

    def __pull_events(self):
        for event in self.__events:
            self.__q.put(event)

    def __process_events(self):
        while not self.__stopped:
            try:
                event = self.__q.get(timeout=self.__event_timeout)
                event = json.loads(event.decode("utf-8"))
            except Empty:
                log.debug("Timed out waiting for event on queue.")
                continue

            for event_handler in self.__event_handlers:
                try:
                    event_handler.handle(event)
                    self.__success_event_count += 1
                except:
                    log.exception("Event handler: '{}' failed to handle event: '{}'".format(
                        type(event_handler).__name__, event))
                    self.__error_event_count += 1

            self.__q.task_done()
            self.__processed_event_count += 1
            log.debug("processed event count: {}".format(self.get_success_count()))

    def set_registry(self, registry):
        self.__reg = registry

    def report_metrics(self, tags):
        self.__reg.gauge(QUEUE_DEPTH_KEY, tags).set(self.get_queue_depth())
        self.__reg.gauge(EVENT_SUCCEEDED_KEY, tags).set(self.get_success_count())
        self.__reg.gauge(EVENT_FAILED_KEY, tags).set(self.get_error_count())
        self.__reg.gauge(EVENT_PROCESSED_KEY, tags).set(self.get_processed_count())




import logging

from titus_isolate.event.constants import ACTION, REBALANCE
from titus_isolate.event.event_handler import EventHandler


class RebalanceEventHandler(EventHandler):

    def __init__(self, workload_manager):
        super().__init__(workload_manager, logging.DEBUG)

    def handle(self, event):
        if not self.__relevant(event):
            return

        self.handling_event(event, "rebalancing workloads")
        self.workload_manager.rebalance()
        self.handled_event(event, "rebalanced workloads")

    def __relevant(self, event):
        if not event[ACTION] == REBALANCE:
            self.ignored_event(event, "not a {} event".format(REBALANCE))
            return False

        return True



import json
import time

from flask import Flask

from titus_isolate.config.constants import TITUS_ISOLATE_BLOCK_SEC, DEFAULT_TITUS_ISOLATE_BLOCK_SEC
from titus_isolate.isolate.detect import get_cross_package_violations, get_shared_core_violations
from titus_isolate.utils import get_config_manager, get_workload_manager, get_event_manager, \
    get_workload_monitor_manager

app = Flask(__name__)


@app.route('/isolate/<workload_id>')
def isolate_workload(workload_id, timeout=None):
    if timeout is None:
        timeout = int(get_config_manager().get(TITUS_ISOLATE_BLOCK_SEC, DEFAULT_TITUS_ISOLATE_BLOCK_SEC))

    deadline = time.time() + timeout
    while time.time() < deadline:
        if get_workload_manager().is_isolated(workload_id):
            return json.dumps({'workload_id': workload_id}), 200, {'ContentType': 'application/json'}
        time.sleep(0.1)

    return json.dumps({'unknown_workload_id': workload_id}), 404, {'ContentType': 'application/json'}


@app.route('/workloads')
def get_workloads():
    workloads = [w.to_dict() for w in get_workload_manager().get_workloads()]
    return json.dumps(workloads)


@app.route('/isolated_workload_ids')
def get_isolated_workload_ids():
    return json.dumps(get_workload_manager().get_isolated_workload_ids())


@app.route('/cpu')
def get_cpu():
    return json.dumps(get_workload_manager().get_cpu().to_dict())


@app.route('/violations')
def get_violations():
    return json.dumps({
        "cross_package": get_cross_package_violations(get_workload_manager().get_cpu()),
        "shared_core": get_shared_core_violations(get_workload_manager().get_cpu())
    })


@app.route('/status')
def get_wm_status():
    return json.dumps({
        "event_manager": {
            "queue_depth": get_event_manager().get_queue_depth(),
            "success_count": get_event_manager().get_success_count(),
            "error_count": get_event_manager().get_error_count(),
            "processed_count": get_event_manager().get_processed_count()
        },
        "workload_manager": {
            "cpu_allocator": get_workload_manager().get_allocator_name(),
            "workload_count": len(get_workload_manager().get_workloads()),
            "isolated_workload_count": len(get_workload_manager().get_isolated_workload_ids()),
            "success_count": get_workload_manager().get_success_count(),
            "error_count": get_workload_manager().get_error_count(),
            "added_count": get_workload_manager().get_added_count(),
            "removed_count": get_workload_manager().get_removed_count()
        }
    })


@app.route('/metrics/raw')
def get_metrics():
    return json.dumps(get_workload_monitor_manager().to_dict())


@app.route('/metrics/cpu_usage/<workload_id>/<seconds>/<granularity>')
def get_cpu_usage(workload_id, seconds, granularity):
    cpu_usage = get_workload_monitor_manager().get_cpu_usage(seconds, granularity)

    if workload_id.lower() == 'all':
        return json.dumps(cpu_usage)

    if workload_id in cpu_usage:
        return json.dumps(cpu_usage[workload_id])

    return json.dumps({'unknown_workload_id': workload_id}), 404, {'ContentType': 'application/json'}



from titus_isolate.event.constants import ACTION
from titus_isolate.event.event_handler import EventHandler
from titus_isolate.event.utils import get_container_name


class FreeEventHandler(EventHandler):
    FREE_ACTIONS = ["die"]

    def __init__(self, workload_manager):
        super().__init__(workload_manager)

    def handle(self, event):
        if not self.__relevant(event):
            return

        workload_id = get_container_name(event)
        self.handling_event(event, "removing workload: '{}'".format(workload_id))
        self.workload_manager.remove_workload(workload_id)
        self.handled_event(event, "removed workload: '{}'".format(workload_id))

    def __relevant(self, event):
        action = event[ACTION].lower()
        if action not in self.FREE_ACTIONS:
            self.ignored_event(event, "action: '{}' should not free cpu resources".format(action))
            return False

        return True



from titus_isolate import log
from titus_isolate.allocate.cpu_allocator import CpuAllocator
from titus_isolate.metrics.constants import FALLBACK_ALLOCATOR_COUNT, FALLBACK_ASSIGN_COUNT, FALLBACK_FREE_COUNT, \
    FALLBACK_REBALANCE_COUNT
from titus_isolate.model.processor.cpu import Cpu


class FallbackCpuAllocator(CpuAllocator):

    def __init__(self, primary_cpu_allocator, secondary_cpu_allocator):
        if primary_cpu_allocator is None:
            raise ValueError("Must be provided a primary cpu allocator.")

        if secondary_cpu_allocator is None:
            raise ValueError("Must be provided a secondary cpu allocator.")

        self.__reg = None

        self.__primary_allocator = primary_cpu_allocator
        self.__secondary_allocator = secondary_cpu_allocator

        self.__secondary_assign_threads_call_count = 0
        self.__secondary_free_threads_call_count = 0
        self.__secondary_rebalance_call_count = 0

        log.debug(
            "Created FallbackCpuAllocator with primary cpu allocator: '{}' and secondary cpu allocator: '{}'".format(
                self.__primary_allocator.__class__.__name__,
                self.__secondary_allocator.__class__.__name__))

    def assign_threads(self, cpu: Cpu, workload_id: str, workloads: dict) -> Cpu:
        try:
            return self.__primary_allocator.assign_threads(cpu, workload_id, workloads)
        except:
            log.exception(
                "Failed to assign threads to workload: '{}' with primary allocator: '{}', falling back to: '{}'".format(
                    workload_id,
                    self.__primary_allocator.__class__.__name__,
                    self.__secondary_allocator.__class__.__name__))
            self.__secondary_assign_threads_call_count += 1
            cpu = self.__secondary_allocator.assign_threads(cpu, workload_id, workloads)
            return cpu

    def free_threads(self, cpu: Cpu, workload_id: str, workloads: dict) -> Cpu:
        try:
            return self.__primary_allocator.free_threads(cpu, workload_id, workloads)
        except:
            log.exception(
                "Failed to free threads for workload: '{}' with primary allocator: '{}', falling back to: '{}'".format(
                    workload_id,
                    self.__primary_allocator.__class__.__name__,
                    self.__secondary_allocator.__class__.__name__))
            self.__secondary_free_threads_call_count += 1
            cpu = self.__secondary_allocator.free_threads(cpu, workload_id, workloads)
            return cpu

    def rebalance(self, cpu: Cpu, workloads: dict) -> Cpu:
        try:
            return self.__primary_allocator.rebalance(cpu, workloads)
        except:
            log.exception(
                "Failed to rebalance workloads: '{}' with primary allocator: '{}', falling back to: '{}'".format(
                    workloads,
                    self.__primary_allocator.__class__.__name__,
                    self.__secondary_allocator.__class__.__name__))
            self.__secondary_rebalance_call_count += 1
            cpu = self.__secondary_allocator.rebalance(cpu, workloads)
            return cpu

    def get_primary_allocator(self) -> CpuAllocator:
        return self.__primary_allocator

    def get_secondary_allocator(self) -> CpuAllocator:
        return self.__secondary_allocator

    def get_fallback_allocator_calls_count(self):
        return self.__secondary_assign_threads_call_count + \
               self.__secondary_free_threads_call_count + \
               self.__secondary_rebalance_call_count

    def set_registry(self, registry):
        self.__reg = registry
        self.__primary_allocator.set_registry(registry)
        self.__secondary_allocator.set_registry(registry)

    def report_metrics(self, tags):
        self.__reg.gauge(FALLBACK_ALLOCATOR_COUNT, tags).set(self.get_fallback_allocator_calls_count())
        self.__reg.gauge(FALLBACK_ASSIGN_COUNT, tags).set(self.__secondary_assign_threads_call_count)
        self.__reg.gauge(FALLBACK_FREE_COUNT, tags).set(self.__secondary_free_threads_call_count)
        self.__reg.gauge(FALLBACK_REBALANCE_COUNT, tags).set(self.__secondary_rebalance_call_count)
        self.__primary_allocator.report_metrics(tags)
        self.__secondary_allocator.report_metrics(tags)

    def str(self):
        return "FallbackCpuAllocator(primary: {}, secondary: {})".format(
            self.__primary_allocator.__class__.__name__,
            self.__secondary_allocator.__class__.__name__)



import calendar
import collections
from datetime import datetime as dt, timedelta as td
from math import ceil
from threading import Lock

import numpy as np

from titus_isolate import log


class WorkloadPerformanceMonitor:

    def __init__(self, metrics_provider, sample_frequency_sec):
        self.__metrics_provider = metrics_provider
        self.__sample_frequency_sec = sample_frequency_sec
        # Maintain buffers for the last hour
        self.__max_buffer_size = ceil(60 * 60 / sample_frequency_sec)
        self.__buffer_lock = Lock()
        self.__timestamps = collections.deque([], self.__max_buffer_size)
        self.__buffers = []

    def get_workload(self):
        return self.__metrics_provider.get_workload()

    def get_buffers(self):
        with self.__buffer_lock:
            return calendar.timegm(dt.utcnow().timetuple()), [calendar.timegm(t.timetuple()) for t in self.__timestamps], [list(e) for e in self.__buffers]
    
    def get_normalized_cpu_usage_last_seconds(self, seconds, agg_granularity_secs=60):
        num_buckets = ceil(seconds / agg_granularity_secs)
        if num_buckets > self.__max_buffer_size:
            raise Exception("Aggregation buffer too small to satisfy query.")
        return WorkloadPerformanceMonitor.normalize_data(*self.get_buffers(), num_buckets, agg_granularity_secs)

    def sample(self):
        cpu_usage_snapshot = self.__metrics_provider.get_cpu_usage()
        if cpu_usage_snapshot is None:
            log.debug("No cpu usage snapshot available for workload: '{}'".format(self.get_workload().get_id()))
            return

        with self.__buffer_lock:
            if len(self.__buffers) == 0:
                self.__buffers = [collections.deque([], self.__max_buffer_size) for _ in range(len(cpu_usage_snapshot.rows))]

            self.__timestamps.append(cpu_usage_snapshot.timestamp)
            for row in cpu_usage_snapshot.rows:
                self.__buffers[row.pu_id].append(int(row.user) + int(row.system))

            log.debug("Took snapshot of metrics for workload: '{}'".format(self.get_workload().get_id()))

    @staticmethod
    def normalize_data(ts_snapshot, timestamps, buffers, num_buckets=60, bucket_size_secs=60):
        proc_time = np.full((num_buckets,), np.nan, dtype=np.float32)

        if len(timestamps) == 0:
            return proc_time

        ts_max = ts_snapshot
        for i in range(num_buckets):
            ts_min = ts_max - bucket_size_secs

            # get slice:
            slice_ts_min = np.searchsorted(timestamps, ts_min)
            slice_ts_max = np.searchsorted(timestamps, ts_max, 'right')
            if slice_ts_max == len(timestamps):
                slice_ts_max -= 1

            ts_max = ts_min

            if slice_ts_min == slice_ts_max:
                continue

            if timestamps[slice_ts_max] < ts_min - bucket_size_secs:
                continue
            # this should be matching Atlas:
            time_diff_ns = (timestamps[slice_ts_max] - timestamps[slice_ts_min]) * 1000000000
            s = 0.0
            for b in buffers: # sum across all cpus
                s += b[slice_ts_max] - b[slice_ts_min]
            if time_diff_ns > 0:
                s /= time_diff_ns
            proc_time[num_buckets - 1 - i] = s

        return proc_time


import time
from threading import Thread, Lock

import schedule

from titus_isolate import log
from titus_isolate.config.agent_property_provider import AgentPropertyProvider
from titus_isolate.config.config_manager import ConfigManager

SCHEDULING_SLEEP_INTERVAL = 1.0

scheduling_lock = Lock()
scheduling_started = False

config_manager_lock = Lock()
__config_manager = None

workload_manager_lock = Lock()
__workload_manager = None

event_manager_lock = Lock()
__event_manager = None

workload_monitor_manager_lock = Lock()
__workload_monitor_manager = None

cpu_usage_predictor_manager_lock = Lock()
__cpu_usage_predictor_manager = None


def get_config_manager(property_provider=AgentPropertyProvider()):
    global __config_manager

    with config_manager_lock:
        if __config_manager is None:
            __config_manager = ConfigManager(property_provider)

        return __config_manager


def set_config_manager(config_manager):
    global __config_manager

    with config_manager_lock:
        __config_manager = config_manager


def get_workload_manager():
    global __workload_manager

    with workload_manager_lock:
        return __workload_manager


def set_workload_manager(workload_manager):
    global __workload_manager

    with workload_manager_lock:
        __workload_manager = workload_manager


def get_event_manager():
    global __event_manager

    with event_manager_lock:
        return __event_manager


def set_event_manager(event_manager):
    global __event_manager

    with event_manager_lock:
        __event_manager = event_manager


def set_workload_monitor_manager(workoad_monitor_manager):
    global __workload_monitor_manager

    with workload_monitor_manager_lock:
        __workload_monitor_manager = workoad_monitor_manager


def get_workload_monitor_manager():
    global __workload_monitor_manager

    with workload_monitor_manager_lock:
        return __workload_monitor_manager


def set_cpu_usage_predictor_manager(cpu_usage_predictor_manager):
    global __cpu_usage_predictor_manager

    with cpu_usage_predictor_manager_lock:
        __cpu_usage_predictor_manager = cpu_usage_predictor_manager


def get_cpu_usage_predictor_manager():
    global __cpu_usage_predictor_manager

    with cpu_usage_predictor_manager_lock:
        return __cpu_usage_predictor_manager


def start_periodic_scheduling():
    global scheduling_started

    with scheduling_lock:
        if scheduling_started:
            return

        worker_thread = Thread(target=__schedule_loop)
        worker_thread.daemon = True
        worker_thread.start()
        scheduling_started = True


def __schedule_loop():
    while True:
        log.debug("Running pending scheduled tasks.")
        schedule.run_pending()

        sleep_time = SCHEDULING_SLEEP_INTERVAL
        if schedule.next_run() is not None:
            sleep_time = schedule.idle_seconds()

        if sleep_time < 0:
            sleep_time = SCHEDULING_SLEEP_INTERVAL

        log.debug("Scheduling thread sleeping for: '{}' seconds".format(sleep_time))
        time.sleep(sleep_time)



from titus_isolate.model.processor.cpu import Cpu
from titus_isolate.monitor.free_thread_provider import FreeThreadProvider


class EmptyFreeThreadProvider(FreeThreadProvider):

    def get_free_threads(self, cpu: Cpu) -> list:
        return [t for t in cpu.get_threads() if len(t.get_workload_ids()) == 0]



import datetime
import json

from titus_isolate.event.constants import WORKLOAD_TYPES, BURST


class Workload:
    def __init__(
            self,
            identifier,
            thread_count,
            mem,
            disk,
            network,
            app_name,
            owner_email,
            image,
            command,
            entrypoint,
            job_type,
            workload_type):

        self.__creation_time = datetime.datetime.utcnow()

        self.__identifier = identifier
        self.__thread_count = int(thread_count)
        self.__mem = mem
        self.__disk = disk
        self.__network = network
        self.__app_name = app_name
        self.__owner_email = owner_email
        self.__image = image
        self.__command = command
        self.__entrypoint = entrypoint
        self.__job_type = job_type
        self.__type = workload_type.lower()

        if self.__thread_count < 0:
            raise ValueError("A workload must request at least 0 threads.")

        if self.__type not in WORKLOAD_TYPES:
            raise ValueError("Unexpected workload type: '{}', acceptable values are: '{}'".format(
                self.__type, WORKLOAD_TYPES))

        if self.__identifier == BURST:
            raise ValueError("The identifier '{}' is reserved".format(BURST))

    def get_id(self):
        return self.__identifier

    def get_thread_count(self):
        return self.__thread_count

    def get_mem(self):
        return self.__mem

    def get_disk(self):
        return self.__disk

    def get_network(self):
        return self.__network

    def get_app_name(self):
        return self.__app_name

    def get_owner_email(self):
        return self.__owner_email

    def get_image(self):
        return self.__image

    def get_command(self):
        return self.__command

    def get_entrypoint(self):
        return self.__entrypoint

    def get_type(self):
        return self.__type

    def get_job_type(self):
        return self.__job_type

    def get_creation_time(self):
        return self.__creation_time

    def set_creation_time(self, creation_time):
        self.__creation_time = creation_time

    def to_dict(self):
        return {
            "creation_time": str(self.__creation_time),
            "id": str(self.get_id()),
            "thread_count": self.get_thread_count(),
            "mem": self.get_mem(),
            "disk": self.get_disk(),
            "network": self.get_network(),
            "app_name": self.get_app_name(),
            "owner_email": self.get_owner_email(),
            "image": self.get_image(),
            "command": self.get_command(),
            "entrypoint": self.get_entrypoint(),
            "job_type": self.get_job_type(),
            "type": self.get_type(),
        }

    def __str__(self):
        return json.dumps(self.to_dict())



from titus_isolate import log
from titus_isolate.event.constants import ACTOR, ATTRIBUTES, NAME, CPU_LABEL_KEY, WORKLOAD_TYPE_LABEL_KEY, \
    REQUIRED_LABELS, MEM_LABEL_KEY, DISK_LABEL_KEY, NETWORK_LABEL_KEY, IMAGE_LABEL_KEY, REPO_DIGESTS, \
    JOB_TYPE_LABEL_KEY, OWNER_EMAIL_LABEL_KEY, APP_NAME_LABEL_KEY, ENTRYPOINT_LABEL_KEY, COMMAND_LABEL_KEY
from titus_isolate.model.workload import Workload


def get_container_name(event):
    return __get_attribute(event, NAME)


def get_cpu(create_event):
    return __get_int_attribute(create_event, CPU_LABEL_KEY)


def get_mem(create_event):
    return __get_int_attribute(create_event, MEM_LABEL_KEY)


def get_disk(create_event):
    return __get_int_attribute(create_event, DISK_LABEL_KEY)


def get_network(create_event):
    return __get_int_attribute(create_event, NETWORK_LABEL_KEY)


def get_app_name(create_event):
    return __get_attribute(create_event, APP_NAME_LABEL_KEY)


def get_owner_email(create_event):
    return __get_attribute(create_event, OWNER_EMAIL_LABEL_KEY)


def get_image(create_event):
    return __get_attribute(create_event, IMAGE_LABEL_KEY)


def get_job_type(create_event):
    return __get_attribute(create_event, JOB_TYPE_LABEL_KEY)


def get_workload_type(create_event):
    return __get_attribute(create_event, WORKLOAD_TYPE_LABEL_KEY)


def get_command(create_event):
    return __get_attribute(create_event, COMMAND_LABEL_KEY)


def get_entrypoint(create_event):
    return __get_attribute(create_event, ENTRYPOINT_LABEL_KEY)


def __get_int_attribute(event, key):
    return int(__get_attribute(event, key, -1))


def __get_attribute(event, key, default=''):
    attributes = event[ACTOR][ATTRIBUTES]
    return __get_value(attributes, key, default)


def __get_value(dictionary, key, default=''):
    return dictionary.get(key, default)


def get_current_workloads(docker_client):
    workloads = []
    for container in docker_client.containers.list():
        workload_id = container.name
        if __has_required_labels(container):
            try:
                labels = container.labels
                cpu = int(__get_value(labels, CPU_LABEL_KEY, -1))
                mem = int(__get_value(labels, MEM_LABEL_KEY, -1))
                disk = int(__get_value(labels, DISK_LABEL_KEY, -1))
                network = int(__get_value(labels, DISK_LABEL_KEY, -1))
                app_name = __get_value(labels, APP_NAME_LABEL_KEY)
                owner_email = __get_value(labels, OWNER_EMAIL_LABEL_KEY)
                command = __get_value(labels, COMMAND_LABEL_KEY)
                entrypoint = __get_value(labels, ENTRYPOINT_LABEL_KEY)
                job_type = __get_value(labels, JOB_TYPE_LABEL_KEY)
                workload_type = __get_value(labels, WORKLOAD_TYPE_LABEL_KEY)
                image = __get_image(container)

                workloads.append(
                    Workload(
                        identifier=workload_id,
                        thread_count=cpu,
                        mem=mem,
                        disk=disk,
                        network=network,
                        app_name=app_name,
                        owner_email=owner_email,
                        image=image,
                        command=command,
                        entrypoint=entrypoint,
                        job_type=job_type,
                        workload_type=workload_type))
                log.info("Found running workload: '{}'".format(workload_id))
            except:
                log.exception("Failed to parse labels for container: '{}'".format(container.name))
        else:
            log.warning("Found running workload: '{}' without expected labels'")

    return workloads


def __get_image(container):
    if REPO_DIGESTS in container.image.attrs:
        repo_digests = container.image.attrs[REPO_DIGESTS]
        if len(repo_digests) > 0:
            return repo_digests[0]

    log.error("Failed to extract image from container: '{}'".format(container.name))
    return ''


def __has_required_labels(container):
    for l in REQUIRED_LABELS:
        if l not in container.labels:
            log.warning("Found running workload: '{}' without expected label: '{}'".format(container.name, l))
            return False

    return True



#!/usr/bin/env python3
import os

from setuptools import setup


install_requires = [
    'click',
    'event',
    'flask'
]

setup(name='titus-isolate',
      description='Isolate containers running on Titus',
      maintainer="Gabriel Hartmann",
      maintainer_email="ghartmann@netflix.com",
      version=os.getenv("TITUS_ISOLATE_VERSION", "0.SNAPSHOT"),
      packages=[
          "titus_isolate",
          "titus_isolate/allocate",
          "titus_isolate/api",
          "titus_isolate/cgroup",
          "titus_isolate/config",
          "titus_isolate/event",
          "titus_isolate/isolate",
          "titus_isolate/metrics",
          "titus_isolate/model",
          "titus_isolate/model/processor",
          "titus_isolate/monitor",
          "titus_isolate/numa",
          "titus_isolate/predict"],
      scripts=["startup/titus-isolate"],
      url="https://github.com/Netflix-Skunkworks/titus-isolate")



import datetime

import boto3 as boto3

from titus_isolate import log
from titus_isolate.config.constants import MODEL_BUCKET_FORMAT_STR, MODEL_BUCKET_PREFIX, \
    DEFAULT_MODEL_BUCKET_PREFIX, MODEL_BUCKET_LEAF, DEFAULT_MODEL_BUCKET_LEAF, MODEL_PREFIX_FORMAT_STR
from titus_isolate.config.utils import get_required_property
from titus_isolate.model.processor.core import Core
from titus_isolate.model.processor.cpu import Cpu
from titus_isolate.model.processor.package import Package
from titus_isolate.model.processor.thread import Thread
from titus_isolate.model.workload import Workload
from titus_isolate.utils import get_config_manager


def get_threads_body(cpu: Cpu, workload_id: str, workloads: dict) -> dict:
    w_list = [w.to_dict() for w in workloads.values()]
    return {
        "cpu": cpu.to_dict(),
        "workload_id": workload_id,
        "workloads": w_list
    }


def get_rebalance_body(cpu: Cpu, workloads: dict) -> dict:
    w_list = [w.to_dict() for w in workloads.values()]
    return {
        "cpu": cpu.to_dict(),
        "workloads": w_list
    }


def parse_cpu(cpu_dict: dict) -> Cpu:
    packages = []
    for p in cpu_dict["packages"]:
        cores = []
        for c in p["cores"]:
            threads = []
            for t in c["threads"]:
                thread = Thread(t["id"])
                for w_id in t["workload_id"]:
                    thread.claim(w_id)
                threads.append(thread)
            cores.append(Core(c["id"], threads))
        packages.append(Package(p["id"], cores))

    return Cpu(packages)


def parse_workload(workload_dict: dict) -> Workload:

    workload = Workload(
        identifier=workload_dict['id'],
        thread_count=workload_dict['thread_count'],
        mem=workload_dict['mem'],
        disk=workload_dict['disk'],
        network=workload_dict['network'],
        app_name=workload_dict['app_name'],
        owner_email=workload_dict['owner_email'],
        image=workload_dict['image'],
        command=workload_dict['command'],
        entrypoint=workload_dict['entrypoint'],
        job_type=workload_dict['job_type'],
        workload_type=workload_dict['type'])

    # Input example:  "2019-03-23 18:03:50.668041"
    creation_time = datetime.datetime.strptime(workload_dict["creation_time"], '%Y-%m-%d %H:%M:%S.%f')
    workload.set_creation_time(creation_time)
    return workload


def get_cpu_model_bucket_name():
    format_str = get_required_property(MODEL_BUCKET_FORMAT_STR)
    if format_str is None:
        return None

    config_manager = get_config_manager()
    region = config_manager.get_region()
    env = config_manager.get_environment()

    return format_str.format(region, env)


def get_cpu_model_prefix_name():
    config_manager = get_config_manager()
    prefix = config_manager.get(MODEL_BUCKET_PREFIX, DEFAULT_MODEL_BUCKET_PREFIX)
    leaf = config_manager.get(MODEL_BUCKET_LEAF, DEFAULT_MODEL_BUCKET_LEAF)

    format_str = get_config_manager().get(MODEL_PREFIX_FORMAT_STR)
    if format_str is None:
        return None

    return format_str.format(prefix, leaf)


def get_cpu_models():
    bucket_name = get_cpu_model_bucket_name()
    if bucket_name is None:
        log.error("Failed to get cpu model bucket name.")
        return None

    prefix_name = get_cpu_model_prefix_name()
    if prefix_name is None:
        log.error("Failed to get cpu model prefix name.")
        return None

    log.info("Getting model metadata from bucket: '{}', prefix: '{}'".format(bucket_name, prefix_name))

    s3_client = boto3.client('s3')
    paginator = s3_client.get_paginator('list_objects')
    pages = paginator.paginate(Bucket=bucket_name, Prefix=prefix_name)

    CONTENTS = 'Contents'
    models = []
    for page in pages:
        if CONTENTS in page:
            for entry in page[CONTENTS]:
                models.append(entry)

    return models


def get_latest_cpu_model():
    models = get_cpu_models()
    if models is None or len(models) == 0:
        return None

    models = sorted(models, key=lambda e: e['LastModified'], reverse=True)
    log.debug("sorted models: {}".format(models))
    return models[0]


def get_cpu_model_file_path():
    return '/var/lib/titus-isolate-cpu-model.bin'


def download_latest_cpu_model(path=get_cpu_model_file_path()):
    log.info("Downloading latest cpu prediction model.")
    latest_model = get_latest_cpu_model()
    if latest_model is None:
        log.error("Failed to download model because no model found.")
        return

    bucket_name = get_cpu_model_bucket_name()
    key = latest_model['Key']
    s3_client = boto3.client('s3')
    s3_client.download_file(bucket_name, key, path)
    log.info("Downloaded latest cpu prediction model: '{}/{}' to: '{}'".format(bucket_name, key, path))



import copy
from threading import Thread, Lock

from titus_isolate import log
from titus_isolate.cgroup.cgroup_manager import CgroupManager
from titus_isolate.cgroup.utils import set_cpuset, wait_for_files, get_cpuset, parse_cpuset
from titus_isolate.config.constants import WAIT_CGROUP_FILE_KEY, WAIT_JSON_FILE_KEY, DEFAULT_WAIT_CGROUP_FILE_SEC, \
    DEFAULT_WAIT_JSON_FILE_SEC
from titus_isolate.metrics.constants import WRITE_CPUSET_FAILED_KEY, WRITE_CPUSET_SUCCEEDED_KEY, ISOLATED_WORKLOAD_COUNT
from titus_isolate.utils import get_config_manager


class FileCgroupManager(CgroupManager):

    def __init__(self):
        self.__reg = None
        self.__lock = Lock()
        self.__write_count = 0
        self.__fail_count = 0
        self.__isolated_workloads = []

    def set_cpuset(self, container_name, thread_ids):
        Thread(target=self.__set_cpuset, args=[container_name, thread_ids]).start()

    def get_cpuset(self, container_name):
        cpuset_str = self.__get_cpuset(container_name)
        if cpuset_str is None:
            return None
        else:
            return parse_cpuset(cpuset_str)

    def release_cpuset(self, container_name):
        with self.__lock:
            if container_name in self.__isolated_workloads:
                self.__isolated_workloads.remove(container_name)

    def get_isolated_workload_ids(self):
        with self.__lock:
            return copy.deepcopy(self.__isolated_workloads)

    def __set_cpuset(self, container_name, thread_ids):
        thread_ids_str = self.__get_thread_ids_str(thread_ids)

        try:
            self.__wait_for_files(container_name)
            log.info("updating workload: '{}' to cpuset.cpus: '{}'".format(container_name, thread_ids_str))
            set_cpuset(container_name, thread_ids_str)
            self.__write_succeeded(container_name)
        except:
            self.__write_failed()
            log.exception("Failed to apply cpuset to threads: '{}' for workload: '{}'".format(
                thread_ids_str, container_name))

    def __get_cpuset(self, container_name: str) -> str:
        try:
            self.__wait_for_files(container_name)
            log.info("getting cpuset.cpus for workload: '{}'".format(container_name))
            return get_cpuset(container_name)
        except:
            log.exception("Failed to read cpuset for workload: '{}'".format(container_name))
            return None

    def __write_succeeded(self, container_name):
        with self.__lock:
            self.__isolated_workloads.append(container_name)
            self.__write_count += 1

    def __write_failed(self):
        with self.__lock:
            self.__fail_count += 1

    @staticmethod
    def __get_thread_ids_str(thread_ids):
        return ",".join([str(t_id) for t_id in thread_ids])

    @staticmethod
    def __wait_for_files(container_name):
        cgroup_file_wait_timeout = int(get_config_manager().get(WAIT_CGROUP_FILE_KEY, DEFAULT_WAIT_CGROUP_FILE_SEC))
        json_file_wait_timeout = int(get_config_manager().get(WAIT_JSON_FILE_KEY, DEFAULT_WAIT_JSON_FILE_SEC))
        wait_for_files(container_name, cgroup_file_wait_timeout, json_file_wait_timeout)

    def set_registry(self, registry):
        self.__reg = registry

    def report_metrics(self, tags):
        self.__reg.gauge(WRITE_CPUSET_SUCCEEDED_KEY, tags).set(self.__write_count)
        self.__reg.gauge(WRITE_CPUSET_FAILED_KEY, tags).set(self.__fail_count)
        self.__reg.gauge(ISOLATED_WORKLOAD_COUNT, tags).set(len(self.get_isolated_workload_ids()))




from titus_isolate.event.constants import BURST, STATIC
from titus_isolate.event.utils import get_container_name, get_cpu, get_mem, get_disk, get_network, get_workload_type, \
    get_image, get_app_name, get_job_type, get_owner_email, get_command, get_entrypoint
from titus_isolate.model.processor.core import Core
from titus_isolate.model.processor.cpu import Cpu
from titus_isolate.model.processor.package import Package
from titus_isolate.model.processor.thread import Thread
from titus_isolate.model.workload import Workload
from titus_isolate.monitor.free_thread_provider import FreeThreadProvider


def get_workload_from_event(event):
    identifier = get_container_name(event)
    cpus = get_cpu(event)
    mem = get_mem(event)
    disk = get_disk(event)
    network = get_network(event)
    app_name = get_app_name(event)
    owner_email = get_owner_email(event)
    image = get_image(event)
    command = get_command(event)
    entrypoint = get_entrypoint(event)
    job_type = get_job_type(event)
    workload_type = get_workload_type(event)

    return Workload(
        identifier=identifier,
        thread_count=cpus,
        mem=mem,
        disk=disk,
        network=network,
        app_name=app_name,
        owner_email=owner_email,
        image=image,
        command=command,
        entrypoint=entrypoint,
        job_type=job_type,
        workload_type=workload_type)


def get_burst_workloads(workloads):
    return get_workloads_by_type(workloads, BURST)


def get_static_workloads(workloads):
    return get_workloads_by_type(workloads, STATIC)


def get_workloads_by_type(workloads, workload_type):
    return [w for w in workloads if w.get_type() == workload_type]


def get_sorted_workloads(workloads):
    return sorted(workloads, key=lambda w: w.get_creation_time())


def release_all_threads(cpu, workloads):
    for w in workloads:
        release_threads(cpu, w.get_id())


def release_threads(cpu, workload_id):
    for t in cpu.get_threads():
        t.free(workload_id)


def update_burst_workloads(cpu, burst_workloads, free_thread_provider):
    free_threads = free_thread_provider.get_free_threads(cpu)
    for t in free_threads:
        for w in burst_workloads:
            t.claim(w.get_id())


def rebalance(cpu: Cpu, workloads: dict, free_thread_provider: FreeThreadProvider) -> Cpu:
    burst_workloads = get_burst_workloads(workloads.values())
    release_all_threads(cpu, burst_workloads)
    update_burst_workloads(cpu, burst_workloads, free_thread_provider)

    return cpu



import requests

from titus_isolate import log
from titus_isolate.config.property_provider import PropertyProvider

PROPERTY_URL_ROOT = 'http://localhost:3002/properties'


class AgentPropertyProvider(PropertyProvider):

    def get(self, key):
        url = self.__get_static_property_url(key)
        response = requests.get(url, headers={"accept": "application/json"})
        if response.status_code != requests.codes.ok:
            log.debug("Failed to retrieve property '{}' with response: '{}'".format(key, response))
            return None

        return response.json()['value']

    @staticmethod
    def __get_static_property_url(key):
        return '{}/property/{}'.format(PROPERTY_URL_ROOT, key)



import copy

from titus_isolate import log


class Thread:
    def __init__(self, processor_id):
        self.__workload_ids = []
        self.__processor_id = int(processor_id)

        if self.__processor_id < 0:
            raise ValueError("Thread processor ids must be non-negative.")

    def get_id(self):
        return self.__processor_id

    def claim(self, workload_id):
        self.__workload_ids.append(workload_id)

    def free(self, workload_id):
        log.debug("Removing workload: '{}' from thread '{}'".format(workload_id, self.get_id()))
        if workload_id in self.__workload_ids:
            self.__workload_ids.remove(workload_id)

    def clear(self):
        log.debug("Removing all workloads: '{}' from thread '{}'".format(self.__workload_ids, self.get_id()))
        self.__workload_ids = []

    def get_workload_ids(self):
        return copy.deepcopy(self.__workload_ids)

    def is_claimed(self):
        return len(self.get_workload_ids()) > 0
